# coding=utf-8
"""Github Storage Backend implementation

This backend stores datasets as GitHub repositories, utilizing Git's built-in
revisions and tags. This implementation is based on GitHub's Web API, and will
not support other Git hosting services.
"""
import json
from typing import Any, Dict, List, Optional, Tuple, Union

from metastore.backend import StorageBackend, exc
from metastore.backend import git_lfs_helpers as lfs_helpers
from metastore.types import Author, PackageRevisionInfo, TagInfo
from metastore.util import is_hex_str

from . import gh_rest_api as gh


class GitHubStorage(StorageBackend):
    """GitHub based metadata storage
    """

    DEFAULT_README = ('# ¯\\_(ツ)_/¯\n'
                      'This is a datapackage repository created by '
                      '[`metastore-lib`](https://github.com/datopian/metastore-lib)')

    DEFAULT_BRANCH = 'master'

    DEFAULT_COMMIT_MESSAGE = 'Datapackage updated'
    DEFAULT_TAG_MESSAGE = 'Tagging revision'

    def __init__(self, github_options, lfs_server_url=None, default_owner=None, default_author=None,
                 default_branch=DEFAULT_BRANCH, default_commit_message=DEFAULT_COMMIT_MESSAGE):
        # type: (Dict[str, Any], Optional[str], Optional[str], Optional[Author], Optional[str], Optional[str]) -> None
        self.gh = gh.Github(**github_options)
        self._lfs_server_url = lfs_server_url
        self._default_owner = default_owner
        self._default_author = default_author
        self._default_branch = default_branch
        self._default_commit_message = default_commit_message
        self._user = None

    def create(self, package_id, metadata, author=None, message=None):
        owner, repo_name = self._parse_id(package_id)
        datapackage = _create_file('datapackage.json', json.dumps(metadata, indent=2))
        files = [datapackage] + self._create_lfs_files(metadata)

        try:
            repo = self._get_owner(owner).create_repo(repo_name)
        except gh.GithubException as e:
            if e.status == 422 and e.data['errors'][0]['message'] == 'name already exists on this account':
                raise exc.Conflict("Datapackage with the same ID already exists")
            raise

        if message is None:
            message = 'Initial datapackage commit'

        try:
            # Create an initial README.md file so we can start using the low-level Git API
            repo.create_file('README.md', 'Initialize data repository', self.DEFAULT_README,
                             branch=self._default_branch)
            head = repo.get_branch(self._default_branch)
            commit = self._create_commit(repo, files, head.commit, author, message)
        except Exception as e:
            try:
                self.delete(package_id)
            except Exception:
                pass
            raise e

        c_author = Author(commit.author.name, commit.author.email)
        return PackageRevisionInfo(package_id, commit.sha, commit.author.date, c_author, message, metadata)

    def fetch(self, package_id, revision_ref=None, repo=None):
        if repo is None:
            repo = self._get_repo(package_id)
        try:
            if not revision_ref:
                ref = repo.get_git_ref('heads/{}'.format(self._default_branch))
                assert ref.object.type == 'commit'
                revision_ref = ref.object.sha
            elif not is_hex_str(revision_ref):
                tag = self.tag_fetch(package_id, revision_ref, repo=repo)
                revision_ref = tag.revision_ref

            # Get the commit pointed by revision_ref
            commit = repo.get_git_commit(revision_ref)

        except gh.UnknownObjectException:
            raise exc.NotFound('Could not find package {}@{}'.format(package_id, revision_ref))

        # Get the blob for datapackage.json in that commit
        try:
            blob = repo.get_contents('datapackage.json', revision_ref)
            datapackage = json.loads(blob.decoded_content)
        except gh.UnknownObjectException:
            raise exc.NotFound("datapackage.json file not found for {}@{}".format(package_id, revision_ref))
        except ValueError:
            raise ValueError("Unable to parse datapackage.json file in {}@{}".format(package_id, revision_ref))

        author = Author(commit.author.name, commit.author.email)
        return PackageRevisionInfo(package_id, commit.sha, commit.author.date, author, commit.message, datapackage)

    def update(self, package_id, metadata, author=None, partial=False, base_revision_ref=None, message=None):
        if message is None:
            message = self._default_commit_message

        repo = self._get_repo(package_id)
        head = repo.get_branch(self._default_branch)

        if partial:
            if base_revision_ref is None:
                base_revision_ref = head.commit.sha
            parent = self.fetch(package_id, base_revision_ref, repo=repo)
            parent.package.update(metadata)
            metadata = parent.package

        datapackage = _create_file('datapackage.json', json.dumps(metadata, indent=2))
        files = [datapackage] + self._create_lfs_files(metadata)
        commit = self._create_commit(repo, files, head.commit, author, message)
        c_author = Author(commit.author.name, commit.author.email)
        return PackageRevisionInfo(package_id, commit.sha, commit.author.date, c_author, message, metadata)

    def delete(self, package_id):
        repo = self._get_repo(package_id)
        repo.delete()

    def revision_list(self, package_id):
        repo = self._get_repo(package_id)
        commits = repo.get_commits(path='datapackage.json')
        revisions = [_commit_to_revinfo(package_id, c) for c in commits]
        return revisions

    def revision_fetch(self, package_id, revision_ref, repo=None):
        return self.fetch(package_id, revision_ref, repo=repo)

    def tag_create(self, package_id, revision_ref, name, author=None, description=None):
        repo = self._get_repo(package_id)
        revision = self.revision_fetch(package_id, revision_ref, repo=repo)
        if description is None:
            description = self.DEFAULT_TAG_MESSAGE

        git_tag = self._create_tag(repo, name, description, revision_ref, author)
        t_author = Author(git_tag.tagger.name, git_tag.tagger.email)
        return TagInfo(package_id, name, git_tag.tagger.date, revision_ref, t_author, revision, description)

    def tag_list(self, package_id):
        repo = self._get_repo(package_id)
        tag_refs = _get_git_matching_refs(repo, 'tags/')
        tags = []
        for ref in tag_refs:
            tags.append(self._tag_ref_to_taginfo(package_id, repo, ref))
        return tags

    def tag_fetch(self, package_id, tag, repo=None):
        if repo is None:
            repo = self._get_repo(package_id)
        try:
            ref = repo.get_git_ref('tags/{}'.format(tag))
        except gh.UnknownObjectException:
            raise exc.NotFound('Could not find tag {} for package {}'.format(tag, package_id))
        return self._tag_ref_to_taginfo(package_id, repo, ref)

    def tag_update(self, package_id, tag, author=None, new_name=None, new_description=None):
        if new_name is None and new_description is None:
            raise ValueError("Expecting at least one of new_name or new_description to be specified")

        repo = self._get_repo(package_id)
        tag_info = self.tag_fetch(package_id, tag, repo=repo)
        name = new_name or tag_info.name
        description = new_description or tag_info.description

        if name == tag_info.name and description == tag_info.description:
            # Nothing to change here
            return tag_info
        elif name != tag_info.name:
            git_tag = self._create_tag(repo, name, description, tag_info.revision_ref, author)
            repo.get_git_ref('tags/{}'.format(tag_info.name)).delete()
        else:
            git_tag = repo.create_git_tag(name, description, tag_info.revision_ref, 'commit')
            repo.get_git_ref('tags/{}'.format(name)).edit(git_tag.sha)

        t_author = Author(git_tag.tagger.name, git_tag.tagger.email)
        return TagInfo(package_id, name, git_tag.tagger.date, tag_info.revision_ref, t_author, tag_info.revision,
                       description)

    def tag_delete(self, package_id, tag):
        repo = self._get_repo(package_id)
        try:
            ref = repo.get_git_ref('tags/{}'.format(tag))
            ref.delete()
        except gh.UnknownObjectException:
            raise exc.NotFound('Could not find tag {} for package {}'.format(tag, package_id))

    def _parse_id(self, package_id):
        # type: (str) -> Tuple(str, str)
        """Verify that the package ID looks like something we can work with and parse
        it into GitHub owner (user or org) and repo name
        """
        if '/' in package_id:
            return tuple(package_id.split('/', 1))
        elif self._default_owner:
            return self._default_owner, package_id
        else:
            raise ValueError('Invalid package ID for the GitHub backend: {}'.format(package_id))

    def _get_owner(self, owner):
        # type: (str) -> Union[gh.AuthenticatedUser, gh.Organization]
        if self._user is None:
            self._user = self.gh.get_user()
        if owner == self._user.login:
            return self._user
        else:
            return self.gh.get_organization(owner)

    def _get_repo(self, package_id):
        # type: (str) -> gh.Repository
        """Get repository object for package_id, validating that it really
        exists
        """
        owner, repo_name = self._parse_id(package_id)
        try:
            return self._get_owner(owner).get_repo(repo_name)
        except gh.UnknownObjectException:
            raise exc.NotFound('Could not find package {}'.format(package_id))

    def _create_commit(self, repo, files, parent_commit, author, message):
        # type: (gh.Repository, List[gh.InputGitTreeElement], gh.Commit, Optional[Author], str) -> gh.GitCommit
        """Create a git Commit
        """
        # Create tree
        tree = repo.create_git_tree(files, parent_commit.commit.tree)
        # Create commit
        author = self._verify_author(author)
        commit = repo.create_git_commit(message, tree, [parent_commit.commit], author=author)
        # Update refs
        ref = repo.get_git_ref('heads/{}'.format(self._default_branch))
        ref.edit(commit.sha)

        return commit

    def _create_tag(self, repo, name, description, revision_ref, author):
        # type: (gh.Repository, str, str, str, Optional[Author]) -> gh.GitTag.GitTag
        """Low level operations for creating a git tag
        """
        author = self._verify_author(author)
        try:
            git_tag = repo.create_git_tag(name, description, revision_ref, 'commit', tagger=author)
            repo.create_git_ref('refs/tags/{}'.format(name), git_tag.sha)
        except gh.GithubException as e:
            if e.status == 422:
                if e.data['message'] == 'Reference already exists':
                    raise exc.Conflict('Tag {} already exists'.format(name))
                # Assume invalid name error
                raise ValueError(e)
            raise

        return git_tag

    def _verify_author(self, author):
        # type: (Optional[Author]) -> Union[gh.InputGitAuthor, gh.NotSet]
        """Check we have an author and return something Git can use to set commit / tag author
        """
        if author and (author.name or author.email):
            return gh.InputGitAuthor(author.name, author.email)
        elif self._default_author:
            return gh.InputGitAuthor(self._default_author.name, self._default_author.email)
        else:
            return gh.NotSet

    def _tag_ref_to_taginfo(self, package_id, repo, ref):
        # type: (str, gh.Repository, gh.GitRef) -> TagInfo
        """Convert a GitRef for a tag into a TagInfo object
        """
        tag_obj = repo.get_git_tag(ref.object.sha)
        revision = self.revision_fetch(package_id, tag_obj.object.sha, repo=repo)
        author = Author(tag_obj.tagger.name, tag_obj.tagger.email)
        return TagInfo(package_id, tag_obj.tag, tag_obj.tagger.date, tag_obj.object.sha, author, revision,
                       tag_obj.message)

    def _create_lfs_files(self, datapackage):
        # type: (Dict[str, Any]) -> List[gh.InputGitTreeElement]
        """Create LFS pointer files and config files, if we need to

        :raise ValueError: If resources with conflicting file names are found
        :raise ValueError: If resources with invalid paths are found
        """
        if not self._lfs_server_url:
            return []

        file_resources = [r for r in datapackage.get('resources', [])
                          if lfs_helpers.is_posix_path_resource(r) and lfs_helpers.has_lfs_attributes(r)]
        if len(file_resources) == 0:
            return []

        paths = set([r['path'] for r in file_resources])
        if len(paths) != len(file_resources):
            raise ValueError('Data package contains resources with conflicting file names')

        for path in paths:
            if path[0] == '/' or '../' in path:
                raise ValueError('Resource path is absolute or contains parent dir references: {}'.format(path))

        files = [_create_file(r['path'], lfs_helpers.create_lfs_pointer_file(r)) for r in file_resources]
        files.append(_create_file('.gitattributes', lfs_helpers.create_git_attributes_file(paths)))
        files.append(_create_file('.lfsconfig', lfs_helpers.create_lfs_config_file(self._lfs_server_url)))

        return files


def _create_file(path, content):
    # type: (str, bytes) -> gh.InputGitTreeElement
    element = gh.InputGitTreeElement(path, '100644', 'blob', content=content)
    return element


def _commit_to_revinfo(package_id, commit):
    # type: (str, gh.Commit) -> PackageRevisionInfo
    """Convert a GitHub Commit object to a PackageRevisionInfo object
    """
    return PackageRevisionInfo(package_id,
                               commit.sha,
                               commit.commit.author.date,
                               Author(commit.commit.author.name, commit.commit.author.email),
                               commit.commit.message)


def _get_git_matching_refs(repo, ref):
    """This is back-ported from PyGithub 1.51 to support Python 2.7 which is no
    longer supported for that version.

    If we ever drop Python 2.7 support, this code is no longer needed and
    :meth:``github.Repository.Repository.get_git_matching_refs`` can be called
    directly.
    """
    if hasattr(repo, 'get_git_matching_refs'):
        return repo.get_git_matching_refs(ref)

    assert isinstance(ref, str), ref
    return gh.PaginatedList.PaginatedList(
        gh.GitRef,
        repo._requester,
        repo.url + "/git/matching-refs/" + ref,
        None,
    )



"""Wrapper around the PyGithub imports to resolve name collisions and clean up calling code a little bit
"""

from github import (AuthenticatedUser, GitCommit, Github, GithubException, GitTag, InputGitTreeElement, Organization,
                    PaginatedList, UnknownObjectException)
from github.Commit import Commit
from github.GithubObject import NotSet
from github.GitRef import GitRef
from github.InputGitAuthor import InputGitAuthor
from github.Repository import Repository

__all__ = ['AuthenticatedUser', 'Commit', 'GitCommit', 'Github', 'GithubException', 'GitTag', 'InputGitTreeElement',
           'Organization', 'PaginatedList', 'Repository', 'UnknownObjectException', 'GitRef', 'NotSet',
           'InputGitAuthor']



"""Tests for the Git LFS helpers module
"""
import pytest

from metastore.backend import git_lfs_helpers


def test_create_git_attributes():
    expected = 'data/* filter=lfs diff=lfs merge=lfs -text\n'
    assert expected == git_lfs_helpers.create_git_attributes_file('data/*')


def test_create_git_attributes_multi_patterns():
    expected = ('data/*.csv filter=lfs diff=lfs merge=lfs -text\n'
                'data/*.xls filter=lfs diff=lfs merge=lfs -text\n')
    assert expected == git_lfs_helpers.create_git_attributes_file(['data/*.csv', 'data/*.xls'])


def test_create_lfs_config_file():
    expected = '[remote "origin"]\n\tlfsurl = https://lfs.example.com/foo/bar'
    assert expected == git_lfs_helpers.create_lfs_config_file('https://lfs.example.com/foo/bar')


def test_create_lfs_config_file_custom_remote_name():
    expected = '[remote "baz-baz"]\n\tlfsurl = https://lfs.example.com/baz/baz'
    assert expected == git_lfs_helpers.create_lfs_config_file('https://lfs.example.com/baz/baz', remote='baz-baz')


def create_lfs_pointer_file():
    resource = {"sha256": "0f1128046248f83dc9b9ab187e16fad0ff596128f1524d05a9a77c4ad932f10a",
                "bytes": 1744,
                "path": "data/resource.csv"}

    expected = ('version https://git-lfs.github.com/spec/v1\n'
                'oid sha256:0f1128046248f83dc9b9ab187e16fad0ff596128f1524d05a9a77c4ad932f10a\n'
                'size 1744\n')

    assert expected == git_lfs_helpers.create_lfs_pointer_file(resource)


@pytest.mark.parametrize('resource', [
    {"bytes": 1744, "path": "data/resource.csv"},
    {"sha256": "0f1128046248f83dc9b9ab187e16fad0ff596128f1524d05a9a77c4ad932f10a"},
    {"bytes": 1744, "sha256": "ef42bab1191da272f13935f78c401e3de0c11afb"},
    {"bytes": 1744, "sha256": "0f112804X248f83dc9b9ab187e16fad0ff596128f1524d05a9a77c4ad932f10a"},
    {"sha256": "0f1128046248f83dc9b9ab187e16fad0ff596128f1524d05a9a77c4ad932f10a", "size": 1744},
])
def create_lfs_pointer_file_invalid_resource(resource):
    with pytest.raises(ValueError):
        git_lfs_helpers.create_lfs_pointer_file(resource)



"""Storage backend exceptions
"""


class StorageBackendError(RuntimeError):
    """Base exception for all storage backend errors
    """
    pass


class Conflict(StorageBackendError):
    """An error representing a conflict between objects

    This may be raised, for example, when trying to modify a package
    which has already been modified in some other manner, or when trying
    to create a package with an ID that already exists.
    """
    pass


class NotFound(StorageBackendError):
    """An error that indicates the user is trying to access an object which
    does not exist or has been deleted from storage
    """
    pass



"""Pyfilesystem based versioned metadata storage

This is useful especially for testing and POC implementations
"""
import hashlib
import json
import re
import uuid
from datetime import datetime
from operator import attrgetter
from typing import Any, Dict, List, Optional

import pytz
import six
from dateutil.parser import isoparse
from fs import open_fs
from fs.errors import DirectoryExists, ResourceNotFound

from ..types import Author, PackageRevisionInfo, TagInfo
from ..util import is_hex_str
from . import StorageBackend, exc


class FilesystemStorage(StorageBackend):
    """Abstract filesystem based storage based on PyFilesystem

    This storage backend is useful mostly in testing, especially with the
    'mem://' file system. You most likely shouldn't be using it in production,
    unless you know exactly what you are doing. This backend does not guarantee
    consistency of storage, and using it in concurrent environments may cause
    issues.

    See the PyFilesystem documentation for types of supported file systems.
    """
    REVISION_DB_FILE = 'revisions.jsonl'

    TAG_NAME_RE = re.compile(r'^[\w\-+.,]+\Z')

    def __init__(self, uri, default_author=None):
        # type: (str, Optional[Author]) -> None
        self._fs = open_fs(uri)
        self._default_author = default_author

    def create(self, package_id, metadata, author=None, message=None):
        try:
            package_dir = self._fs.makedirs(_get_package_path(package_id))
        except DirectoryExists:
            raise exc.Conflict("Package with id {} already exists".format(package_id))

        revision = _make_revision_id()
        with self._fs.lock():
            with package_dir.open(revision, 'wb') as f:
                f.write(json.dumps(metadata).encode('utf8'))
            rev_info = self._log_revision(package_id, revision, author, message)
        rev_info.package = metadata
        return rev_info

    def fetch(self, package_id, revision_ref=None):
        if not revision_ref:
            # Get the latest revision
            revision = self.revision_list(package_id)[0]
            revision_ref = revision.revision
        elif not is_hex_str(revision_ref, chars=32):
            tag = self.tag_fetch(package_id, revision_ref)
            revision_ref = tag.revision_ref
            revision = tag.revision
        else:
            revision = self.revision_fetch(package_id, revision_ref)

        package_path = u'{}/{}'.format(_get_package_path(package_id), revision_ref)
        try:
            with self._fs.open(package_path, 'r') as f:
                revision.package = json.load(f)
        except (ResourceNotFound, ValueError):
            raise exc.NotFound('Could not find package {}@{}', package_id, revision_ref)

        return revision

    def update(self, package_id, metadata, author=None, partial=False, base_revision_ref=None, message=None):
        current = self.fetch(package_id, base_revision_ref)

        if partial:
            current.package.update(metadata)
            metadata = current.package

        with self._fs.lock():
            revision = _make_revision_id()
            with self._fs.open(u'{}/{}'.format(_get_package_path(package_id), revision), 'wb') as f:
                f.write(json.dumps(metadata).encode('utf8'))
            rev_info = self._log_revision(package_id, revision, author, message)
        rev_info.package = metadata
        return rev_info

    def delete(self, package_id):
        path = _get_package_path(package_id)
        try:
            self._fs.removetree(path)
        except ResourceNotFound:
            raise exc.NotFound('Could not find package {}', package_id)

    def revision_list(self, package_id):
        try:
            revisions = self._get_revisions(package_id)
        except ResourceNotFound:
            raise exc.NotFound('Could not find package {}', package_id)
        return revisions

    def revision_fetch(self, package_id, revision_ref):
        try:
            revision = self._get_revision(package_id, revision_ref)
        except ResourceNotFound:
            raise exc.NotFound('Could not find package {}', package_id)
        if revision is None:
            raise exc.NotFound('Could not find package {}@{}', package_id, revision_ref)
        return revision

    def tag_create(self, package_id, revision_ref, name, author=None, description=None):
        if not self._validate_tag_name(name):
            raise ValueError("Invalid tag name: {}".format(name))
        revision = self.revision_fetch(package_id, revision_ref)
        return self._log_tag(revision, name, author, description)

    def tag_list(self, package_id):
        tags = []
        tag_dir = self._open_tag_dir(package_id)
        if tag_dir is None:
            return []

        for tag_name in tag_dir.listdir('.'):
            with tag_dir.open(tag_name, 'r') as f:
                tag_line = json.load(f)
            tags.append(_tag_file_to_taginfo(package_id, tag_name, tag_line))

        return sorted(tags, key=attrgetter('created'))

    def tag_fetch(self, package_id, tag):
        if not self._validate_tag_name(tag):
            raise ValueError("Invalid tag name: {}".format(tag))

        tag_info = self._get_tag(package_id, tag)
        if not tag_info:
            raise exc.NotFound('Could not find tag {} for package {}', tag, package_id)

        tag_info.revision = self.revision_fetch(package_id, tag_info.revision_ref)
        return tag_info

    def tag_update(self, package_id, tag, author=None, new_name=None, new_description=None):
        if new_name is None and new_description is None:
            raise ValueError("Expecting at least one of new_name or new_description to be specified")
        elif new_name and not self._validate_tag_name(new_name):
            raise ValueError("Invalid tag name: {}".format(new_name))

        tag_info = self.tag_fetch(package_id, tag)
        name = new_name or tag_info.name
        description = new_description or tag_info.description or None
        overwrite = tag_info.name == name

        with self._fs.lock():
            tag_info = self._log_tag(tag_info.revision, name, author, description, allow_overwrite=overwrite)
            if not overwrite:
                self.tag_delete(package_id, tag)

        return tag_info

    def tag_delete(self, package_id, tag):
        with self._fs.lock():
            tags_dir = self._open_tag_dir(package_id)
            try:
                tags_dir.remove(tag)
            except ResourceNotFound:
                raise exc.NotFound('Could not find tag {} for package {}', tag, package_id)

    def _log_revision(self, package_id, revision, author, message=None):
        # type: (str, str, Optional[Author], Optional[str]) -> PackageRevisionInfo
        """Log a revision
        """
        db_file = u'{}/{}'.format(_get_package_path(package_id), self.REVISION_DB_FILE)
        now = datetime.now(tz=pytz.utc).isoformat()
        author = self._verify_author(author)

        with self._fs.open(db_file, 'ab') as f:
            rev_info = {"revision": revision,
                        "created": now,
                        "description": message,
                        "author_name": author.name,
                        "author_email": author.email}
            f.write('{}\n'.format(json.dumps(rev_info)).encode('utf8'))
        return PackageRevisionInfo(package_id, revision, now, author, message)

    def _get_revisions(self, package_id):
        # type: (str) -> List[PackageRevisionInfo]
        """Get list of revisions from DB file
        """
        db_file = u'{}/{}'.format(_get_package_path(package_id), self.REVISION_DB_FILE)
        with self._fs.open(db_file, 'r') as f:
            lines = [json.loads(line) for line in f]
            revisions = [_parse_rev_log(package_id, line) for line in reversed(lines)]
        return revisions

    def _get_revision(self, package_id, revision):
        # type: (str, str) -> PackageRevisionInfo
        """Get a specific revision from the revisions DB file

        If not found, will return None
        """
        db_file = u'{}/{}'.format(_get_package_path(package_id), self.REVISION_DB_FILE)
        with self._fs.open(db_file, 'r') as f:
            for line in f:
                rev_data = json.loads(line)
                if rev_data['revision'] == revision:
                    return _parse_rev_log(package_id, rev_data)

    def _validate_tag_name(self, name):
        # type: (str) -> bool
        """Validate a tag name
        """
        return bool(self.TAG_NAME_RE.match(name))

    def _log_tag(self, revision, tag_name, author, tag_description, allow_overwrite=False):
        # type: (PackageRevisionInfo, str, Optional[Author], str, bool) -> TagInfo
        """Log a new tag for an existing revision
        """
        tags_path = u'{}/{}'.format(_get_package_path(revision.package_id), 'tags')
        now = datetime.now(tz=pytz.utc)
        tags_dir = self._fs.makedirs(tags_path, recreate=True)
        author = self._verify_author(author)

        with tags_dir.lock():
            if not allow_overwrite and tags_dir.exists(tag_name):
                raise exc.Conflict('Tag already exists: {}'.format(tag_name))

            with tags_dir.open(tag_name, 'wb') as f:
                tag_info = {"created": now.isoformat(),
                            "revision": revision.revision,
                            "description": tag_description,
                            "author_name": author.name,
                            "author_email": author.email}
                f.write(json.dumps(tag_info).encode('utf8'))
        return TagInfo(revision.package_id, tag_name, now, revision.revision, author, revision,
                       description=tag_description)

    def _get_tag(self, package_id, tag_name):
        # type: (str, str) -> Optional[TagInfo]
        """Get a specific tag from the tags DB file

        If not found, will return None
        """
        try:
            tag_dir = self._open_tag_dir(package_id)
            if not tag_dir:
                return None
            with tag_dir.open(tag_name, 'r') as f:
                line = json.load(f)
        except ResourceNotFound:
            return None

        return _tag_file_to_taginfo(package_id, tag_name, line)

    def _open_tag_dir(self, package_id):
        """Open a tag directory and return it
        """
        try:
            package_dir = self._fs.opendir(_get_package_path(package_id))
        except ResourceNotFound:
            raise exc.NotFound('Could not find package {}', package_id)

        try:
            return package_dir.opendir('tags')
        except ResourceNotFound:
            return None

    def _verify_author(self, author):
        # type: (Optional[Author]) -> Author
        """Verify that we have a valid author object
        """
        if author:
            return author
        if self._default_author:
            return self._default_author
        return Author(None, None)


def _get_package_path(package_id):
    # type: (str) -> six.text_type
    """Create a package path
    """
    return u'/p/{}'.format(hashlib.sha256(package_id.encode('utf8')).hexdigest())


def _make_revision_id():
    # type: () -> str
    """Generate a random unique revision ID
    """
    return uuid.uuid4().hex


def _parse_rev_log(package_id, rev_data):
    # type: (str, Dict[str, Any]) -> PackageRevisionInfo
    """Parse a line from the revision log and return a RevisionInfo object
    """
    author = _get_author(rev_data)
    return PackageRevisionInfo(package_id=package_id,
                               revision=rev_data['revision'],
                               created=isoparse(rev_data['created']),
                               author=author,
                               description=rev_data['description'])


def _tag_file_to_taginfo(package_id, tag_name, tag_data):
    # type: (str, str, Dict[str, Any]) -> TagInfo
    author = _get_author(tag_data)
    return TagInfo(package_id=package_id,
                   name=tag_name,
                   created=isoparse(tag_data['created']),
                   revision_ref=tag_data['revision'],
                   description=tag_data['description'],
                   author=author)


def _get_author(record):
    # type: (Dict[str, Any]) -> Optional[Author]
    if 'author_name' in record or 'author_email' in record:
        author = Author(name=record.get('author_name'),
                        email=record.get('author_email'))
    else:
        author = None
    return author



import importlib
from typing import Callable, Optional


def get_callable(callable_str, base_package=None):
    # type: (str, Optional[str]) -> Callable
    """Get a callable function / class constructor from a string of the form
    `package.subpackage.module:callable`

    >>> type(get_callable('os.path:basename')).__name__
    'function'

    >>> type(get_callable('basename', 'os.path')).__name__
    'function'
    """
    if ':' in callable_str:
        module_name, callable_name = callable_str.split(':', 1)
        module = importlib.import_module(module_name, base_package)
    elif base_package:
        module = importlib.import_module(base_package)
        callable_name = callable_str
    else:
        raise ValueError("Expecting base_package to be set if only class name is provided")

    return getattr(module, callable_name)  # type: ignore


def is_hex_str(value, chars=40):
    # type: (str, int) -> bool
    """Check if a string is a hex-only string of exactly :param:`chars` characters length.

    This is useful to verify that a string contains a valid SHA, MD5 or UUID-like value.

    >>> is_hex_str('0f1128046248f83dc9b9ab187e16fad0ff596128f1524d05a9a77c4ad932f10a', 64)
    True

    >>> is_hex_str('0f1128046248f83dc9b9ab187e16fad0ff596128f1524d05a9a77c4ad932f10a', 32)
    False

    >>> is_hex_str('0f1128046248f83dc9b9ab187e1xfad0ff596128f1524d05a9a77c4ad932f10a', 64)
    False

    >>> is_hex_str('ef42bab1191da272f13935f78c401e3de0c11afb')
    True

    >>> is_hex_str('ef42bab1191da272f13935f78c401e3de0c11afb'.upper())
    True

    >>> is_hex_str('ef42bab1191da272f13935f78c401e3de0c11afb', 64)
    False

    >>> is_hex_str('ef42bab1191da272f13935.78c401e3de0c11afb')
    False
    """
    if len(value) != chars:
        return False
    try:
        int(value, 16)
    except ValueError:
        return False
    return True



"""Tests for the filysystem based storage backend
"""
import pytest

from metastore.backend.filesystem import FilesystemStorage

from . import CommonBackendTestSuite


@pytest.fixture()
def backend():
    return FilesystemStorage('mem://')


class TestFilesystemBackend(CommonBackendTestSuite):
    pass



"""Base library tests
"""
from metastore.backend import create_metastore
from metastore.backend.filesystem import FilesystemStorage
from metastore.backend.github import GitHubStorage


def test_factory_instantiates_filesystem_backend():
    backend = create_metastore('filesystem', dict(uri='mem://'))
    assert isinstance(backend, FilesystemStorage)


def test_factory_instantiates_github_backend():
    backend = create_metastore('github', {"github_options": {}})
    assert isinstance(backend, GitHubStorage)



"""Data classes for packages & related models
"""
from datetime import datetime
from typing import Any, Dict, Optional

# Python type definition for datapackage.json dicts
# This may be further specified in the future (esp. if Python 2.x support is dropped)
DataPackageType = Dict[str, Any]


# Data classes. If this was Python 3.7+, we could have used dataclasses here ;)
class Author(object):
    """Revision / tag author information
    """
    name = None  # type: Optional[str]
    email = None  # type: Optional[str]

    def __init__(self, name=None, email=None):
        self.name = name
        self.email = email

    def __eq__(self, other):
        return _compare_attributes(self, other, ('name', 'email'))

    def __repr__(self):
        return '<Author {}>'.format(self.email)


class PackageRevisionInfo(object):
    """Revision information
    """
    package_id = None  # type: str
    revision = None  # type: str
    created = None  # type: datetime
    author = None  # type: Optional[Author]
    description = None  # type: Optional[str]
    package = None  # type: Optional[DataPackageType]

    def __init__(self, package_id, revision, created, author=None, description=None, package=None):
        self.package_id = package_id
        self.revision = revision
        self.created = created
        self.author = author
        self.description = description
        self.package = package

    def __eq__(self, other):
        return _compare_attributes(self, other, ('package_id', 'revision'))


class TagInfo(object):
    """Tag information
    """
    package_id = None  # type: str
    name = None  # type: str
    created = None  # type: datetime
    revision_ref = None  # type: str
    author = None  # type: Optional[Author]
    revision = None  # type: Optional[PackageRevisionInfo]
    description = None  # type: Optional[str]

    def __init__(self, package_id, name, created, revision_ref, author=None, revision=None, description=None):
        self.package_id = package_id
        self.name = name
        self.created = created
        self.revision_ref = revision_ref
        self.author = author
        self.revision = revision
        self.description = description

    def __eq__(self, other):
        return _compare_attributes(self, other, ('package_id', 'name'))


def _compare_attributes(obj, other, key_attributes):
    """Object comparison helper
    """
    if not isinstance(obj, other.__class__):
        return NotImplemented

    try:
        return all(getattr(obj, a) == getattr(obj, a) for a in key_attributes)
    except AttributeError:
        return False



"""Unit tests for the GitHub backend

NOTE: these tests mock actual requests / responses to GitHub
"""
import logging
import os

import pytest
from github import GithubException

from metastore.backend.github import GitHubStorage

from . import CommonBackendTestSuite, create_test_datapackage


class CleaningGitHubStorage(GitHubStorage):
    """A wrapper around the GitHub storage class that cleans up after itself
    """

    def __init__(self, *args, **kwargs):
        super(CleaningGitHubStorage, self).__init__(*args, **kwargs)
        self._packages = set()
        self._log = logging.getLogger(__name__)

    def create(self, package_id, metadata, author=None, message=None):
        package = super(CleaningGitHubStorage, self).create(package_id, metadata, author, message)
        self._packages.add(package.package_id)
        return package

    def cleanup__(self):
        for pkg in self._packages:
            try:
                self.delete(pkg)
            except (GithubException, RuntimeError):
                self._log.warning("Failed cleaning up GitHub repo after test: %s", pkg)
        self._packages = set()


@pytest.fixture()
def backend():
    backend = CleaningGitHubStorage(github_options={"login_or_token": os.environ.get('GITHUB_TOKEN')},
                                    default_owner=os.environ.get('GITHUB_OWNER'))
    try:
        yield backend
    finally:
        backend.cleanup__()


@pytest.fixture()
def lfs_backend():
    backend = CleaningGitHubStorage(github_options={"login_or_token": os.environ.get('GITHUB_TOKEN')},
                                    default_owner=os.environ.get('GITHUB_OWNER'),
                                    lfs_server_url='https://lfs.example.com/foo/bar')
    try:
        yield backend
    finally:
        backend.cleanup__()


def create_test_lfs_package(name, **kwargs):
    if 'resources' in kwargs:
        return create_test_datapackage(name, **kwargs)

    pkg = create_test_datapackage(name, **kwargs)
    pkg['resources'][0].update({
        "sha256": '0f1128046248f83dc9b9ab187e16fad0ff596128f1524d05a9a77c4ad932f10a',
        "bytes": 1744
    })
    return pkg


@pytest.fixture(scope='module')
def vcr_config():
    have_github_access = bool(os.environ.get('GITHUB_TOKEN') and os.environ.get('GITHUB_OWNER'))
    force_record = bool(os.environ.get('DISABLE_VCR'))
    if have_github_access:
        if force_record:
            mode = 'all'
        else:
            mode = 'once'
    else:
        mode = 'none'

    return {
        "filter_headers": [
            ('authorization', 'fake-authz-header')
        ],
        "record_mode": mode
    }


@pytest.mark.vcr()
class TestGitHubBackend(CommonBackendTestSuite):

    ID_PREFIX = 'test__'

    # GitHub has different rules for valid tag names

    @pytest.mark.parametrize('name', [
        'with space',
        'with\n',
        '',
    ])
    def test_tag_create_invalid_names(self, name, backend):
        p1 = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        with pytest.raises(ValueError):
            backend.tag_create(p1.package_id, p1.revision, name, description="Invalid tag name")

    def test_datapackage_create_fetch_with_lfs(self, lfs_backend):
        """Test we can create and fetch packages with LFS configured
        """
        p1_info = lfs_backend.create(self.dataset_id('mydataset'), create_test_lfs_package('mydataset'))

        assert p1_info.package_id == self.dataset_id('mydataset')
        assert p1_info.revision is not None
        assert p1_info.package['resources'][0]['path'] == 'data/myresource.csv'

        p2_info = lfs_backend.fetch(self.dataset_id('mydataset'))
        assert p1_info.package == p2_info.package
        assert p2_info.package['resources'][0]['bytes'] == 1744

    def test_datapackage_create_with_lfs_conflicting_resources(self, lfs_backend):
        """Test the backend does not accept conflicting LFS resources
        """
        datapkg = create_test_lfs_package('mydataset')
        datapkg['resources'].append({
            "path": "data/myresource.csv",
            "bytes": 1515,
            "sha256": "08419486253228102a04995a0376ffdaec0bf1dbaf9cff3669f34d29ad483a02",
        })

        with pytest.raises(ValueError):
            lfs_backend.create(self.dataset_id('mydataset'), datapkg)

    @pytest.mark.parametrize('path', [
        '/foo/someresource.csv',
        'foo/../../someresource.csv',
        '../foo/someresource.csv'
    ])
    def test_datapackage_create_with_lfs_invalid_resources(self, lfs_backend, path):
        """Test the backend does not accept conflicting LFS resources
        """
        datapkg = create_test_lfs_package('mydataset')
        datapkg['resources'][0].update({
            "path": path,
            "bytes": 1515,
            "sha256": "08419486253228102a04995a0376ffdaec0bf1dbaf9cff3669f34d29ad483a02",
        })

        with pytest.raises(ValueError):
            lfs_backend.create(self.dataset_id('mydataset'), datapkg)



"""Backend namespace
"""
from typing import Any, Dict, List, Optional

from metastore.types import Author, DataPackageType, PackageRevisionInfo, TagInfo
from metastore.util import get_callable

BACKEND_CLASSES = {'github': 'metastore.backend.github:GitHubStorage',
                   'filesystem': 'metastore.backend.filesystem:FilesystemStorage'}


class StorageBackend(object):
    """Abstract interface for storage backend classes
    """
    def create(self, package_id, metadata, author=None, message=None):
        # type: (str, DataPackageType, Optional[Author], Optional[str]) -> PackageRevisionInfo
        """Create a new data package
        """
        raise NotImplementedError("This method is not implemented for this backend")

    def fetch(self, package_id, revision_ref=None):
        # type: (str, Optional[str]) -> PackageRevisionInfo
        """Fetch a data package, potentially at a given revision / tag

        :raises metastore.backend.exc.NotFound: Raised if the package or referenced revision do not exist
        """
        raise NotImplementedError("This method is not implemented for this backend")

    def update(self, package_id, metadata, author=None, partial=False, message=None, base_revision_ref=None):
        # type: (str, DataPackageType, Optional[Author], bool, Optional[str], Optional[str]) -> PackageRevisionInfo
        """Update or partial update (patch) a data package
        """
        raise NotImplementedError("This method is not implemented for this backend")

    def delete(self, package_id):
        # type: (str) -> None
        """Delete a data package

        :raises metastore.backend.exc.NotFound: Raised if the package does not exist
        """
        raise NotImplementedError("This method is not implemented for this backend")

    def revision_list(self, package_id):
        # type: (str) -> List[PackageRevisionInfo]
        """Get list of revisions for a data package

        Note that when listing, :attr:`metastore.types.PackageRevisionInfo.package`
        will not be set in the return value. If you need to access the actual package
        metadata and not just revision information, fetch the package data by calling
        :meth:`fetch`.

        :raises metastore.backend.exc.NotFound: Raised if the package does not exist
        """
        raise NotImplementedError("This method is not implemented for this backend")

    def revision_fetch(self, package_id, revision_ref):
        # type: (str, str) -> PackageRevisionInfo
        """Get info about a specific revision of a data package

        NOTE: this does not fetch the data package metadata itself, just info about a revision of it

        :raises metastore.backend.exc.NotFound: Raised if the package or revision do not exist
        """
        raise NotImplementedError("This method is not implemented for this backend")

    def tag_create(self, package_id, revision_ref, name, author=None, description=None):
        # type: (str, str, str, Optional[Author], Optional[str]) -> TagInfo
        """Create a tag (named reference to a revision of a data package)

        :raises metastore.backend.exc.NotFound: Raised if the package or referenced revision do not exist
        """
        raise NotImplementedError("This method is not implemented for this backend")

    def tag_list(self, package_id):
        # type: (str) -> List[TagInfo]
        """Get list of tags for a package

        :raises metastore.backend.exc.NotFound: Raised if the package does not exist
        """
        raise NotImplementedError("This method is not implemented for this backend")

    def tag_fetch(self, package_id, tag):
        # type: (str, str) -> TagInfo
        """Get info about a specific tag of a data package

        NOTE: this does not fetch the data package metadata itself, just info about a revision of it

        :raises metastore.backend.exc.NotFound: Raised if the package or tag do not exist
        """
        raise NotImplementedError("This method is not implemented for this backend")

    def tag_update(self, package_id, tag, author=None, new_name=None, new_description=None):
        # type: (str, str, Optional[Author], Optional[str], Optional[str]) -> TagInfo
        """Modify existing tag name or description

        :raises metastore.backend.exc.NotFound: Raised if the package or tag do not exist
        :raises metastore.backend.exc.Conflict: Raised if the tag name is already in use by another tag
        """
        raise NotImplementedError("This method is not implemented for this backend")

    def tag_delete(self, package_id, tag):
        # type: (str, str) -> None
        """Delete an existing

        :raises metastore.backend.exc.NotFound: Raised if the package or tag do not exist
        """
        raise NotImplementedError("This method is not implemented for this backend")


def create_metastore(backend_type, options):
    # type: (str, Dict[str, Any]) -> StorageBackend
    """Factory for storage backends

    Note that this will import the right backend module only as needed
    """
    if ':' not in backend_type:
        try:
            backend_type = BACKEND_CLASSES[backend_type]
        except KeyError:
            raise ValueError("Unknown backend type: {}".format(type))

    backend = get_callable(backend_type)
    return backend(**options)  # type: ignore



from os import path
from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.md')) as f:
    long_description = f.read()

setup(
    name='metastore-lib',
    packages=find_packages(exclude='tests'),
    version=open('VERSION').read(),
    description='A library for abstracting versioned metadata storage for data packages',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Shahar Evron',
    author_email='shahar.evron@datopian.com',
    install_requires=[
        'pytz',
        'six',
        'typing',
        'typing-extensions',
        'python-dateutil>=2.7.0',
        'fs',
        'pygithub',
    ],
    package_data={}
)



from typing import Any, Dict

import pytest

from metastore.backend.filesystem import exc
from metastore.types import Author


def create_test_datapackage(name, **kwargs):
    # type: (str, ...) -> Dict[str, Any]
    """Create a datapackage.json structure for testing purposes
    """
    package = {"name": name,
               "resources": [
                   {"path": "data/myresource.csv"}
               ]}
    package.update(kwargs)
    return package


class CommonBackendTestSuite(object):

    ID_PREFIX = 'myorg/'

    @classmethod
    def dataset_id(cls, name):
        return '{}{}'.format(cls.ID_PREFIX, name)

    def test_datapackage_create_fetch(self, backend):
        p1_info = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))

        assert p1_info.package_id == self.dataset_id('mydataset')
        assert p1_info.revision is not None
        assert p1_info.package['resources'][0]['path'] == 'data/myresource.csv'

        p2_info = backend.fetch(self.dataset_id('mydataset'))
        assert p1_info.package == p2_info.package

    def test_datapackage_create_name_conflict(self, backend):
        package_info = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        assert package_info.package_id == self.dataset_id('mydataset')

        with pytest.raises(exc.Conflict):
            backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))

    def test_datapackage_fetch_not_found(self, backend):
        with pytest.raises(exc.NotFound):
            backend.fetch(self.dataset_id('mydataset'))

    def test_datapackage_update_partial(self, backend):
        rev1_info = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        rev2_info = backend.update(self.dataset_id('mydataset'), {"type": "csv"}, partial=True)
        assert rev1_info.package_id == rev2_info.package_id
        assert rev1_info.revision != rev2_info.revision

        package_info = backend.fetch(self.dataset_id('mydataset'))
        assert package_info.revision == rev2_info.revision
        assert package_info.package['type'] == 'csv'
        assert package_info.package['name'] == 'mydataset'

    def test_datapackage_update_full(self, backend):
        backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset', type='xls'))
        backend.update(self.dataset_id('mydataset'), {"name": "my-data-set"})

        package_info = backend.fetch(self.dataset_id('mydataset'))
        assert package_info.package['name'] == 'my-data-set'
        assert 'type' not in package_info.package

    def test_datapackage_fetch_revision(self, backend):
        pkg = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        r1 = backend.update(pkg.package_id, {"type": "csv"}, partial=True)
        backend.update(pkg.package_id, {"author": "jimbob@example.com"}, partial=True)

        pkg_info = backend.fetch(pkg.package_id, revision_ref=r1.revision)
        assert pkg_info.revision == r1.revision
        assert 'author' not in pkg_info.package
        assert 'type' in pkg_info.package

    def test_datapackage_delete(self, backend):
        package_info = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        assert package_info.package_id == self.dataset_id('mydataset')

        backend.delete(package_info.package_id)
        with pytest.raises(exc.NotFound):
            backend.fetch(self.dataset_id('mydataset'))

    def test_datapackage_delete_non_existing(self, backend):
        with pytest.raises(exc.NotFound):
            backend.delete(self.dataset_id('someid'))

    def test_revision_list_one_revision(self, backend):
        p1 = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        revs = backend.revision_list(self.dataset_id('mydataset'))
        assert len(revs) == 1
        assert revs[0].revision == p1.revision

    def test_revision_list_multiple_revisions(self, backend):
        p1 = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        p2 = backend.update(self.dataset_id('mydataset'), {"type": "csv"}, partial=True,
                            message="Set type to csv")
        p3 = backend.update(self.dataset_id('mydataset'), {"type": "xls"}, partial=True,
                            message="Set type to xls")

        revs = backend.revision_list(self.dataset_id('mydataset'))
        assert len(revs) == 3
        assert revs[2].revision == p1.revision
        assert revs[1].revision == p2.revision
        assert revs[1].description == "Set type to csv"
        assert revs[0].revision == p3.revision
        assert revs[0].description == "Set type to xls"

    def test_revision_list_multiple_revisions_different_packages(self, backend):
        p1 = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        p2 = backend.create(self.dataset_id('otherdataset'), create_test_datapackage('otherdataset'))
        p3 = backend.update(self.dataset_id('mydataset'), {"type": "csv"}, partial=True,
                            message="Set type to csv")
        p4 = backend.update(self.dataset_id('otherdataset'), {"type": "xls"}, partial=True,
                            message="Set type to xls")

        revs = backend.revision_list(self.dataset_id('mydataset'))
        assert len(revs) == 2
        assert revs[1].revision == p1.revision
        assert revs[0].revision == p3.revision
        assert revs[0].description == "Set type to csv"

        revs = backend.revision_list(self.dataset_id('otherdataset'))
        assert len(revs) == 2
        assert revs[1].revision == p2.revision
        assert revs[0].revision == p4.revision
        assert revs[0].description == "Set type to xls"

    def test_revision_fetch(self, backend):
        backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        p2 = backend.update(self.dataset_id('mydataset'), {"type": "csv"}, partial=True,
                            message="Set type to csv")

        rev = backend.revision_fetch(p2.package_id, p2.revision)
        assert p2.revision == rev.revision
        assert "Set type to csv" == rev.description

    def test_revision_fetch_no_package(self, backend):
        with pytest.raises(exc.NotFound):
            backend.revision_fetch(self.dataset_id('mydataset'), '123123123123123')

    def test_revision_fetch_no_revision(self, backend):
        p1 = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        rev = backend.revision_fetch(p1.package_id, p1.revision)
        assert p1.revision == rev.revision
        with pytest.raises(exc.NotFound):
            backend.revision_fetch(p1.package_id, '123123123123123123')

    def test_revision_list_multiple_authors(self, backend):
        author1 = Author(name='Bob Example', email='bob@example.com')
        author2 = Author(name='Bilbo Baggins', email='bilbo@shire.net')
        backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'), author=author1)
        backend.update(self.dataset_id('mydataset'), {"type": "csv"}, partial=True, author=author2,
                       message="Set type to csv")
        backend.update(self.dataset_id('mydataset'), {"type": "xls"}, partial=True, author=author1,
                       message="Set type to xls")

        revs = backend.revision_list(self.dataset_id('mydataset'))
        assert len(revs) == 3
        assert revs[2].author == author1
        assert revs[1].author == author2
        assert revs[0].author == author1
        assert revs[0].author.email == 'bob@example.com'
        assert revs[0].author.name == 'Bob Example'

    def test_tag_create_fetch(self, backend):
        p1 = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        t1 = backend.tag_create(p1.package_id, p1.revision, 'version-1.0', description='My nice little tag')
        t2 = backend.tag_fetch(p1.package_id, 'version-1.0')
        assert t2.revision == t1.revision
        assert t2.name == 'version-1.0'
        assert t2.description == 'My nice little tag'

    @pytest.mark.parametrize('name', [
        'with space',
        'with!',
        'with\n',
        '',
        'foo\tbar',
    ])
    def test_tag_create_invalid_names(self, name, backend):
        p1 = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        with pytest.raises(ValueError):
            backend.tag_create(p1.package_id, p1.revision, name, description="Invalid tag name")

    def test_tag_create_fetch_substring_names(self, backend):
        p1 = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        r1 = backend.tag_create(p1.package_id, p1.revision, 'version-1.0', description="My nice little tag")
        r2 = backend.tag_create(p1.package_id, p1.revision, 'version-1.0.0', description="This is a different tag")
        t1 = backend.tag_fetch(p1.package_id, 'version-1.0')
        assert t1.revision == r1.revision
        assert t1.name == 'version-1.0'
        t2 = backend.tag_fetch(p1.package_id, 'version-1.0.0')
        assert t2.revision == r2.revision
        assert t2.name == 'version-1.0.0'

    def test_tag_create_existing_name(self, backend):
        p1 = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        backend.tag_create(p1.package_id, p1.revision, 'version-1.0', description="My nice little tag")
        p2 = backend.update(self.dataset_id('mydataset'), create_test_datapackage('mydataset', type='csv'))
        with pytest.raises(exc.Conflict):
            backend.tag_create(p1.package_id, p2.revision, 'version-1.0', description="Next Tag with Same Name")

    def test_tag_fetch_no_tag(self, backend):
        p1 = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        with pytest.raises(exc.NotFound):
            backend.tag_fetch(p1.package_id, 'version-1.0')

    def test_tag_fetch_no_package(self, backend):
        p1 = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        backend.tag_create(p1.package_id, p1.revision, 'version-1.0', description="My nice little tag")
        with pytest.raises(exc.NotFound):
            backend.tag_fetch(self.dataset_id('otherdataset'), 'version-1.0')

    def test_tag_list(self, backend):
        p1 = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        backend.tag_create(p1.package_id, p1.revision, 'version-1.0', description="First version")
        p2 = backend.update(self.dataset_id('mydataset'), create_test_datapackage('mydataset', type='csv'))
        backend.tag_create(p1.package_id, p2.revision, 'version-1.1', description="Second version")
        p3 = backend.create(self.dataset_id('otherdataset'), create_test_datapackage('otherdataset'))
        backend.tag_create(p3.package_id, p3.revision, 'version-1.0', description="First version")

        tags = backend.tag_list(self.dataset_id('mydataset'))
        assert len(tags) == 2
        assert tags[0].name == 'version-1.0'
        assert tags[0].revision_ref == p1.revision
        assert tags[1].name == 'version-1.1'
        assert tags[1].revision_ref == p2.revision

    def test_tag_list_no_tags(self, backend):
        p1 = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        backend.tag_create(p1.package_id, p1.revision, 'version-1.0', description="First version")
        backend.create(self.dataset_id('otherdataset'), create_test_datapackage('otherdataset'))

        tags = backend.tag_list(self.dataset_id('otherdataset'))
        assert 0 == len(tags)

    def test_tag_list_no_package(self, backend):
        with pytest.raises(exc.NotFound):
            backend.tag_list(self.dataset_id('otherdataset'))

    def test_tag_list_authors(self, backend):
        author1 = Author(name='Bob Example', email='bob@example.com')
        author2 = Author(name='Bilbo Baggins', email='bilbo@shire.net')
        author3 = Author(name='Frodo Baggins', email='frodo@shire.net')
        p1 = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'), author=author1)
        backend.tag_create(p1.package_id, p1.revision, 'version-1.0', description="First version", author=author2)
        p2 = backend.update(self.dataset_id('mydataset'), create_test_datapackage('mydataset', type='csv'),
                            author=author1)
        backend.tag_create(p1.package_id, p2.revision, 'version-1.1', description="Second version", author=author3)

        tags = backend.tag_list(self.dataset_id('mydataset'))
        assert len(tags) == 2
        assert tags[0].author == author2
        assert tags[1].author == author3
        assert tags[1].author.email == 'frodo@shire.net'
        assert tags[1].author.name == 'Frodo Baggins'

    def test_tag_update(self, backend):
        p1 = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        t1 = backend.tag_create(p1.package_id, p1.revision, 'version-1.0', description="My nice little tag")
        t2 = backend.tag_update(p1.package_id, t1.name, new_name='v-1.0', new_description='My newer better tag')

        with pytest.raises(exc.NotFound):
            backend.fetch(p1.package_id, revision_ref=t1.name)

        p2 = backend.fetch(p1.package_id, revision_ref=t2.name)
        assert p2 == p1

        tags = backend.tag_list(p1.package_id)
        assert len(tags) == 1
        assert tags[0].description == 'My newer better tag'

    def test_tag_update_desc_only(self, backend):
        p1 = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        t1 = backend.tag_create(p1.package_id, p1.revision, 'version-1.0', description="My nice little tag")
        backend.tag_update(p1.package_id, t1.name, new_description='My newer better tag')

        p2 = backend.fetch(p1.package_id, revision_ref=t1.name)
        assert p2 == p1

        tags = backend.tag_list(p1.package_id)
        assert len(tags) == 1
        assert tags[0].description == 'My newer better tag'

    def test_tag_update_name_only(self, backend):
        p1 = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        t1 = backend.tag_create(p1.package_id, p1.revision, 'version-1.0', description="My nice little tag")
        t2 = backend.tag_update(p1.package_id, t1.name, new_name='v-1.0')

        with pytest.raises(exc.NotFound):
            backend.fetch(p1.package_id, revision_ref=t1.name)

        p2 = backend.fetch(p1.package_id, revision_ref=t2.name)
        assert p2 == p1

        tags = backend.tag_list(p1.package_id)
        assert len(tags) == 1
        assert tags[0].description == 'My nice little tag'
        assert tags[0].name == 'v-1.0'

    def test_tag_update_no_change(self, backend):
        p1 = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        t1 = backend.tag_create(p1.package_id, p1.revision, 'version-1.0', description="My nice little tag")
        with pytest.raises(ValueError):
            backend.tag_update(p1.package_id, t1.name)

    def test_tag_update_existing_name(self, backend):
        p1 = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        t1 = backend.tag_create(p1.package_id, p1.revision, 'version-1.0', description="My nice little tag")
        p2 = backend.update(self.dataset_id('mydataset'), create_test_datapackage('mydataset', type='csv'))
        t2 = backend.tag_create(p1.package_id, p2.revision, 'version-1.1', description="My next version tag")

        with pytest.raises(exc.Conflict):
            backend.tag_update(p1.package_id, t1.name, new_name=t2.name)

        tags = backend.tag_list(p1.package_id)
        assert len(tags) == 2

    def test_tag_update_no_tag(self, backend):
        p1 = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        with pytest.raises(exc.NotFound):
            backend.tag_update(p1.package_id, 'version-1.0', new_name='v-1.0')

    def test_tag_update_no_package(self, backend):
        with pytest.raises(exc.NotFound):
            backend.tag_update(self.dataset_id('mydataset'), 'version-1.0', new_name='v-1.0')

    def test_tag_update_invalid_name(self, backend):
        p1 = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        t1 = backend.tag_create(p1.package_id, p1.revision, 'version-1.0', description="My nice little tag")

        with pytest.raises(ValueError):
            backend.tag_update(p1.package_id, t1.name, new_name='ver 1.0')

    def test_tag_delete(self, backend):
        p1 = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        backend.tag_create(p1.package_id, p1.revision, 'version-1.0', description="First version")
        p2 = backend.update(self.dataset_id('mydataset'), create_test_datapackage('mydataset', type='csv'))
        backend.tag_create(p1.package_id, p2.revision, 'version-1.1', description="Second version")
        p3 = backend.create(self.dataset_id('otherdataset'), create_test_datapackage('otherdataset'))
        backend.tag_create(p3.package_id, p3.revision, 'version-1.0', description="First version")

        tags = backend.tag_list(self.dataset_id('mydataset'))
        assert len(tags) == 2

        backend.tag_delete(p1.package_id, 'version-1.0')

        tags = backend.tag_list(self.dataset_id('mydataset'))
        assert 1 == len(tags)
        assert 1 == len(backend.tag_list(self.dataset_id('otherdataset')))
        assert tags[0].name == 'version-1.1'

    def test_tag_delete_no_tag(self, backend):
        p1 = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        backend.tag_create(p1.package_id, p1.revision, 'version-1.0', description="First version")

        with pytest.raises(exc.NotFound):
            backend.tag_delete(p1.package_id, 'version-1.1')

    def test_tag_delete_no_package(self, backend):
        p1 = backend.create(self.dataset_id('mydataset'), create_test_datapackage('mydataset'))
        backend.tag_create(p1.package_id, p1.revision, 'version-1.0', description="First version")

        with pytest.raises(exc.NotFound):
            backend.tag_delete(self.dataset_id('otherdataset'), 'version-1.0')



"""Git LFS Helpers used by all Git based backends that support Git-LFS
"""
from typing import Any, Dict, Iterable, Union

from six import string_types

from ..util import is_hex_str


def is_posix_path_resource(resource):
    # type: (Dict[str, Any]) -> bool
    """Tell if a resource represents a POSIX-path (i.e. local file) resource

    >>> is_posix_path_resource({"path": "data/file.csv"})
    True

    >>> is_posix_path_resource({"path": "a file with some stange name"})
    True

    >>> is_posix_path_resource({"path": "http://example.com/my-resource"})
    False

    >>> is_posix_path_resource({"data": "some-inline-data"})
    False

    >>> is_posix_path_resource({"path": ["file1.csv", "file2.csv"]})
    False
    """
    if 'path' not in resource:
        return False

    if not isinstance(resource['path'], str):
        return False

    if resource['path'][:7] in {'http://', 'https:/'}:
        return False

    return True


def has_lfs_attributes(resource):
    # type: (Dict[str, Any]) -> bool
    """Tell if a resource has the attributes required for an LFS-stored resource

    >>> has_lfs_attributes({"path": "data.csv", "bytes": 1234, "sha256": "someshavalue"})
    True

    >>> has_lfs_attributes({"path": "data.csv", "size": 1234, "sha256": "someshavalue"})
    False

    >>> has_lfs_attributes({"path": "data.csv", "sha256": "someshavalue"})
    False

    >>> has_lfs_attributes({"path": "data.csv", "bytes": 1234})
    False

    >>> has_lfs_attributes({"path": "data.csv"})
    False
    """
    if not isinstance(resource.get('sha256'), str):
        return False

    if not isinstance(resource.get('bytes'), int):
        return False

    return True


def create_git_attributes_file(track_files):
    # type: (Union[str, Iterable[str]]) -> str
    """Create the contents of a .gitattributes file as required by Git LFS
    """
    if isinstance(track_files, string_types):
        track_files = [track_files]

    lines = ['{} filter=lfs diff=lfs merge=lfs -text\n'.format(p) for p in track_files]
    return ''.join(lines)


def create_lfs_config_file(lfs_server_url, remote='origin'):
    # type: (str, str) -> str
    """Create contents of .lfsconfig file
    """
    return '[remote "{remote}"]\n\tlfsurl = {lfsurl}'.format(remote=remote, lfsurl=lfs_server_url)


def create_lfs_pointer_file(resource):
    # type: (Dict[str, Any]) -> str
    """Create contents for LFS pointer file
    """
    if not is_hex_str(resource['sha256'], chars=64):
        raise ValueError('Resource sha256 value does not seem to be a valid sha256 hex string')

    return ('version https://git-lfs.github.com/spec/v1\n'
            'oid sha256:{}\n'
            'size {}\n').format(resource['sha256'], resource['bytes'])
